
// Matrix product

//#include <stdlib.h>
//#include <stdio.h>
#include <memory.h>
#include "MPIinclude.h"
#include "globals.h"
#include "threads.h"
#include "ProcTree.h"
#include "Procs.h"
#include "Range.h"
#include "DAD.h"
#include "schedule.h"
#include "MessSchedule.h"
#include "TreeSchedule.h"
#include "Reduce.h"
#include "Remap.h"
#include "matmul.h"

void boolMatmul(DAD* res,  int* resDat,
                DAD* src1, int* src1Dat, DAD* src2, int* src2Dat) {
  Group  p = res->grp() ;
  Range  x = res->rng(0), y = res->rng(1) ;
  Map    s = res->map(0), t = res->map(1) ;

  Range x1 = src1->rng(0) ;
  Range y2 = src2->rng(1) ;
  Range z1 = src1->rng(1), z2 = src2->rng(0) ;

  // Require:
  //   x1.size() == x.size(), y2.size() == y.size(), z1.size() == z2.size()

  const int bandWidth = MAX_TEMPORARY_SIZE /
                        (sizeof(int) * x.volume() + sizeof(int) * y.volume()) ;

  LocBlocksIndex i(x), j(y) ;

  if(p.member()) {
    for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
      const int m = i.count ;
      //const int resStep0  = s.step(i.sub_stp) ;
      const int resStep0  = s.str() * i.sub_stp ;

      const int resOffset0 = s.offset(i) ;

      for(j.beginLocBlk() ; j.test() ; j.nextLocBlk()) {
	const int n = j.count ;
	//const int resStep1  = t.step(j.sub_stp) ;
	const int resStep1  = t.str() * j.sub_stp ;

	const int resOffset1 = resOffset0 + t.offset(j) ;

	int* resBlock = resDat  + resOffset1 ;

	for(int il = 0 ; il < m ; il++) {
	  int* resPtr = resBlock  ;
	  for(int jl = 0 ; jl < n ; jl++) {
	    *resPtr = 0 ;
	    resPtr += resStep1  ;
	  }
	  resBlock += resStep0  ;
	}
      }
    }
  }

  int base = 0, residue = z1.size() ;
  while(residue) {
    const int b = bandWidth < residue ? bandWidth : residue ;
    CollapsedRange zb(b) ;

    Map tmp1_maps [2], tmp2_maps [2], sec1_maps [2], sec2_maps [2] ;
    Map s1, t2 ;
    int t1, s2 ;

    // Set up array descriptors for temporaries

    int size ;

    // Array2<int> tmp1(x, zb, p) ;

    size = 1 ;
    tmp1_maps [0] = Map(x, size) ;
    size *= x.volume() ;
    tmp1_maps [1] = Map(zb, size) ;
    size *= zb.volume() ;

    DAD tmp1(2, p, tmp1_maps) ;
    int* tmp1Dat = new int [size] ;

    s1 = tmp1.map(0) ;
    t1 = tmp1.str(1) ;

    // Array2<int> tmp2(zb, y, p) ;

    size = 1 ;
    tmp2_maps [0] = Map(zb, size) ;
    size *= zb.volume() ;
    tmp2_maps [1] = Map(y, size) ;
    size *= y.volume() ;

    DAD tmp2(2, p, tmp2_maps) ;
    int* tmp2Dat = new int [size] ;

    s2 = tmp2.str(0) ;
    t2 = tmp2.map(1) ;

    // Set up array descriptors for sections

    // Section2<int> sec1 = src1.sect(x1, Trip(base, ceil))) ;

    sec1_maps [0] = src1->map(0) ;
    sec1_maps [1] = Map(z1.subrng(b, base), src1->str(1)) ;
    DAD sec1(2, src1->group, sec1_maps) ;

    // Section2<int> sec2 = src2.sect(Trip(base, ceil), y2)) ;

    sec2_maps [0] = Map(z2.subrng(b, base), src2->str(0)) ;
    sec2_maps [1] = src2->map(1) ;
    DAD sec2(2, src2->group, sec2_maps) ;

    // Copy sections of source to temporaries

    Remap remap1(&tmp1, &sec1, sizeof(int)) ;
    remap1.execute(tmp1Dat, src1Dat) ;

    Remap remap2(&tmp2, &sec2, sizeof(int)) ;
    remap2.execute(tmp2Dat, src2Dat) ;

    // Local multiplication

    if(p.member()) {

      for(i.beginLocBlk() ; i.test() ; i.nextLocBlk()) {
	const int m = i.count ;
        //const int resStep0  = s.step(i.sub_stp) ;
        const int resStep0  = s.str() * i.sub_stp ;
        //const int tmp1Step0 = s1.step(i.sub_stp) ;
        const int tmp1Step0 = s1.str() * i.sub_stp ;

	const int resOffset0 = s.offset(i) ;
	const int tmp1Offset = s1.offset(i) ;

	for(j.beginLocBlk() ; j.test() ; j.nextLocBlk()) {
	  const int n = j.count ;
          //const int resStep1  = t.step(j.sub_stp) ;
          const int resStep1  = t.str() * j.sub_stp ;
          //const int tmp2Step1 = t2.step(j.sub_stp) ;
          const int tmp2Step1 = t2.str() * j.sub_stp ;

	  const int resOffset1 = resOffset0 + t.offset(j) ;
	  const int tmp2Offset = t2.offset(j) ;

	  int* resBlock  = resDat  + resOffset1 ;
	  int* tmp1Block = tmp1Dat + tmp1Offset ;
	  int* tmp2Block = tmp2Dat + tmp2Offset ;

	  // Should really use an optimized sequential matrix multiplier here.

	  for(int il = 0 ; il < m ; il++) {
	    int* resPtr   = resBlock  ;
	    int* tmp2Ptr1 = tmp2Block ;
	    for(int jl = 0 ; jl < n ; jl++) {
	      int* tmp1Ptr  = tmp1Block ;
	      int* tmp2Ptr2 = tmp2Ptr1 ;
	      for(int k = 0 ; k < b ; k++) {
		*resPtr  = *resPtr || (*tmp1Ptr && *tmp2Ptr2) ;
		tmp1Ptr  += t1 ;
		tmp2Ptr2 += s2 ;
	      }
	      resPtr   += resStep1  ;
	      tmp2Ptr1 += tmp2Step1 ;
	    }
	    resBlock  += resStep0  ;
	    tmp1Block += tmp1Step0 ;
	  }
	}
      }
    }

    delete [] tmp2Dat ;
    delete [] tmp1Dat ;

    base    += b ;
    residue -= b ;
  }
}

// -------------------------------------------------------------
// Things to do
//

